/* cons.c   by Michael Thorpe   2005-08-10 */

#include <stdlib.h>
#include <string.h>
#include "lisp.h"

/* This function doesn't inc/dec the refcnt for s or list */
static obj *evallist(scope *s,obj *list) {
	obj *a,*c,*d;

	if(list->type != &cons_objtype)
		return(evaluate(incscoperef(s),incref(list)));
	c=newobj(&cons_objtype);
	if(c) {
		a=evaluate(incscoperef(s),incref(list->value.c.a));
		if(a) {
			d=evallist(s,list->value.c.d);
			if(d) {
				c->value.c.a=a;
				c->value.c.d=d;
				return(c);
			}
			decref(a);
		}
		c->value.c.a=incref(&nilobj);
		c->value.c.d=incref(&nilobj);
		decref(c);
	}
	return(0);
}

static obj *cons_evaluate(scope *s,obj *o) {
	obj *func,*args;

	func=evaluate(incscoperef(s),incref(o->value.c.a));
	if(!func) {
		decscoperef(s);
		decref(o);
		return(0);
	}
	args=evallist(s,o->value.c.d);
	if(!args) {
		decref(func);
		decscoperef(s);
		decref(o);
		return(0);
	}
	decref(o);
	return((func->type->applyfunc)(s,func,args));
}

static obj *cons_tostring(obj *in) {
	obj *out=0,*tmp,**a;
	char *s;
	unsigned long i,l,n;
	int halfopen=0;

	for(i=0,tmp=in;tmp->type==&cons_objtype;tmp=tmp->value.c.d)
		i++;
	a=(obj **)malloc((i+1)*sizeof(obj *));
	if(!a) {
		decref(in);
		return(throwoom());
	}
	l=1; /*  2 (for the open/close parens)
	        -1 (for the "extra" space char on item 0) */
	n=0;
	tmp=in;
	while(tmp->type==&cons_objtype) {
		a[n]=(tmp->value.c.a->type->tostring)(incref(tmp->value.c.a));
		if(!a[n])
			goto fail;
		if(a[n]->type != &string_objtype) {
			throwtypeerror(a[n]);
			goto fail;
		}
		l += 1+a[n]->value.s.len; /* the 1 is for the separating space */
		n++;
		tmp=tmp->value.c.d;
	}
	if(tmp->type != &nil_objtype) {
		a[n]=(tmp->type->tostring)(incref(tmp));
		if(!a[n])
			goto fail;
		if(a[n]->type != &string_objtype) {
			throwtypeerror(a[n]);
			goto fail;
		}
		l += 3+a[n]->value.s.len; /* the 3 is for the " . " */
		n++;
		halfopen=1;
	}
	out=real_new_string(0,l);
	if(!out)
		goto fail;
	s=out->value.s.data;
	s[0]='(';
	l=1;
	for(i=0;i<n;i++) {
		if(i) {
			if(i+1 == n && halfopen) {
				s[l++]=' ';
				s[l++]='.';
			}
			s[l++]=' ';
		}
		memcpy(&s[l],a[i]->value.s.data,a[i]->value.s.len);
		l += a[i]->value.s.len;
	}
	s[l]=')';
fail:
	while(0<n--)
		decref(a[n]);
	free(a);
	decref(in);
	return(out);
}

static void decref_cons(obj *o) {
	decref(o->value.c.a);
	decref(o->value.c.d);
}

obj *newlist(unsigned long length) {
	obj *c,*o;

	c=incref(&nilobj);
	while(length) {
		o=newobj(&cons_objtype);
		if(!o) {
			decref(c);
			return(0);
		}
		o->value.c.a=incref(&nilobj);
		o->value.c.d=c;
		c=o;
		length--;
	}
	return(c);
}

objtype cons_objtype={
	name: "cons",
	tostring: cons_tostring,
	dropobj: decref_cons,
	evaluate: cons_evaluate
};

