/* main.c   by Michael Thorpe   2005-07-25 */

#include <stdio.h>
#include "lisp.h"

int tryit() {
	obj *s,*o;
	FILE *f;

	f=fopen("preload.lisp","r");
	if(!f) {
		perror("fopen");
		return(1);
	}
	s=newobj(&stream_objtype);
	if(!s) {
		fclose(f);
		return(1);
	}
	s->value.stream.readable=1;
	s->value.stream.writable=0;
	s->value.stream.closable=1;
	s->value.stream.f=f;
	if(store_object("#preload.lisp",incref(s)))
		return(1);
	while(1) {
		o=read_object(incref(s));
		if(!o && real_eqv(exception,eoferror)) {
			decref(exception);
			decref(exceptionvalue);
			exception=incref(&nilobj);
			exceptionvalue=incref(&nilobj);
			break;
		}
		if(!o) {
			decref(s);
			fclose(f);
			return(1);
		}
		o=evaluate(incscoperef(globalscope),o);
		if(!o) {
			decref(s);
			fclose(f);
			return(1);
		}
		decref(o);
	}
	decref(s);
	return(0);
}

int main() {
#ifdef DEBUG
	obj *o;
#endif

	if(scope_init()
	|| builtin_init()
	|| quote_init()
	|| reader_init()
	|| stream_init()
	|| string_init()
	|| throw_init()
	|| util_init()) {
		fprintf(stderr,"got bad return value from an init!\n");
		return(1);
	}
	if(tryit()) {
		fprintf(stderr,"\ngot bad return value from tryit!\n");
#ifdef DEBUG
		fprintf(stderr,"EXCEPTION: ");
		o=writeobj(incref(stderrobj),incref(exception));
		decref(o);
		fprintf(stderr,"\nEXCEPTION-VALUE: ");
		o=writeobj(incref(stderrobj),incref(exceptionvalue));
		decref(o);
		fprintf(stderr,"\n");
#endif
		return(1);
	}
	decscoperef(globalscope);
	util_exit();
#ifdef DEBUG
	fflush(stdout);
	if(total_scopes)
		fprintf(stderr,"Total scopes left over: %lu\n",total_scopes);
	if(total_objects != known_objects) {
		fprintf(stderr,"Total objects left over: %lu (%lu known)\n",total_objects,known_objects);
		show_remaining_objects();
	}
	return((total_objects != known_objects)||total_scopes);
#else
	return(0);
#endif
}

