/* util.c   by Michael Thorpe   2005-08-10 */

#include <stdlib.h>
#include <string.h>
#include "lisp.h"

static obj *freelist=0;
#ifdef DEBUG
unsigned long known_objects=0,total_objects=0;
static obj *firstobj=0;

obj *real_incref(obj *o,const char *file,int line) {
#if DEBUG > 2
	fprintf(stderr,"%s:%d: incref(%p)\n",file,line,o);
#endif
	if(o->rcnf.refcount>10000000)
		*(char *)0=0;
	o->rcnf.refcount++;
	return(o);
}

void real_decref(obj *o,const char *file,int line) {
#if DEBUG > 2
	fprintf(stderr,"%s:%d: decref(%p)\n",file,line,o);
#endif
	if(!--o->rcnf.refcount)
		drop_object(o);
}
#endif

void drop_object(obj *o) {
#ifdef DEBUG
#if DEBUG > 1
	fprintf(stderr,"drop_object(%p)\n",o);
#endif
	if(o->type->dropobj)
		(o->type->dropobj)(o);
	total_objects--;
	if(o->prev)
		o->prev->next=o->next;
	else
		firstobj=o->next;
	if(o->next)
		o->next->prev=o->prev;
	memset(o,'o',sizeof(obj));
#else
	o->rcnf.nextfree=freelist;
	freelist=o;
#endif
}

static obj *objtypeof(obj *o) {
	obj *t;

	t=incref(o->type->nameobj);
	decref(o);
	return(t);
}

static obj *rawpassthrough(scope *s,obj *o) {
	decscoperef(s);
	return(o);
}

static obj *rawtypeerror(scope *s,obj *func,obj *args) {
	decscoperef(s);
	decref(args);
	return(throwtypeerror(func));
}

obj *evaluate(scope *scope,obj *o) {
	return((o->type->evaluate)(scope,o));
}

obj *newobj(objtype *type) {
	obj *o;

	if(freelist) {
		o=freelist;
		freelist=o->rcnf.nextfree;
		if(o->type->dropobj)
			(o->type->dropobj)(o);
	} else {
		o=(obj *)calloc(1,sizeof(obj));
		if(!o)
			return(throwoom());
	}
	o->type=type;
	o->rcnf.refcount=1;
#ifdef DEBUG
#if DEBUG > 1
	fprintf(stderr,"newobj(%s)=%p\n",type->name,o);
#endif
	total_objects++;
	o->prev=0;
	o->next=firstobj;
	if(firstobj)
		firstobj->prev=o;
	firstobj=o;
#endif
	return(o);
}

int register_objtype(objtype *type) {
	if(!type->nameobj)
		type->nameobj=newword(type->name,strlen(type->name));
	if(!type->nameobj)
		return(1);
#ifdef DEBUG
	known_objects++;
#endif
	if(!type->tostring)
		return(1);
	if(!type->evaluate)
		type->evaluate=rawpassthrough;
	if(!type->applyfunc)
		type->applyfunc=rawtypeerror;
	return(0);
}

#ifdef DEBUG
void show_remaining_objects() {
	obj *o;

	for(o=firstobj;o;o=o->next) {
		if((obj *)0x6f6f6f6f==o->next) {
			fprintf(stderr,"%p: dropped object found!\n",o);
			break;
		}
		fprintf(stderr,"%p: <%lu:%s:",o,o->rcnf.refcount,o->type->name);
		if(&word_objtype==o->type)
			fprintf(stderr,"%p=\"%.*s\"",o->value.s.data,(int)o->value.s.len,o->value.s.data);
		else
			decref(writeobj(incref(stderrobj),incref(o)));
		fprintf(stderr,">\n");
	}
}
#endif

int util_init() {
	if(store_builtin_s1("evaluate",evaluate)
	|| store_builtin_ns1("object-type",objtypeof)
/* Should these should be handled in the per-type init functions? */
	|| register_objtype(&cons_objtype)
	|| register_objtype(&float_objtype)
	|| register_objtype(&func_objtype)
	|| register_objtype(&integer_objtype)
	|| register_objtype(&nil_objtype)
	|| register_objtype(&quote_objtype)
	|| register_objtype(&rawfunc_objtype)
	|| register_objtype(&stream_objtype)
	|| register_objtype(&string_objtype)
	|| register_objtype(&word_objtype))
		return(1);
	return(0);
}

void util_exit() {
	obj *o;

	while(freelist) {
		o=freelist;
		freelist=o->rcnf.nextfree;
		if(o->type->dropobj)
			(o->type->dropobj)(o);
		free(o);
	}
}
